import re

# TODO: shouldn't be the case that filters are always classes

class PassThroughFilter(object):

    def __call__(self, line, buffer):
        return (line, buffer)


class RegexpFilter(object):
    
    def __init__(self, regex=None, invert=False):
        self._invert = invert
        if not regex:
            raise ValueError("regex argument can't be blank")
        self._re = re.compile(regex)
        
    def __call__(self, line, buffer):
        found = bool(self._re.search(line))
        if found and (not self._invert):
            return (line, buffer)
        elif (not found) and self._invert:
            return (line, buffer)
        else:
            return (None, None)

        
class WordsFilter(RegexpFilter):
    
    def __init__(self, words, invert=False):
        regex = ''
        if not hasattr(words, '__iter__'):
            esc = map(re.escape, words)
            regex = '\b(%s)\b' % '|'.join(esc)
        else:
            regex = re.escape(words)
        RegexpFilter.__init__(self, regex, invert)



from parser import parser
from lex import lexer

class Config(object):

    def __init__(self, watches):
        self.watching = watches

    @classmethod
    def from_file(cls, file):
        config = None
        env = {}
        with open(file) as f:
            inp = f.read()
            cn = parser.parse(inp, lexer=lexer)
            config = cn.evaluate(env)
        return config



import ply.lex as lex

reserved_words = {
    'route': 'ROUTE',
    'watch': 'WATCH',
    'filter': 'FILTER',
}

tokens = [
    'COMMA',
    'EQUALS',
    'AT',
    'LCURLY',
    'RCURLY',
    'NUMBER',
    'STRING',
    'BOOL',
    'IMPORT_PATH',
    'IDENTIFIER',
] + reserved_words.values()

t_AT = r'@'
t_EQUALS = r'='
t_COMMA = r','
t_LCURLY = r'\{'
t_RCURLY = r'\}'
t_ignore_COMMENT = r'\#.*'

def t_NUMBER(t):
    r'\d+'
    t.value = int(t.value)    
    return t

def t_STRING(t):
    r'"(\\[nrt"\\]|[^\\\"])*"'
    replacements = [
        ('\\n', '\n'),
        ('\\r', '\r'),
        ('\\t', '\t'),
        ('\\"', '"'),
        ('\\\\', '\\'),
        ]
    for s, r in replacements:
        t.value = t.value.replace(s, r)
    t.value = t.value[1:-1]
    return t

def t_IMPORT_PATH(t):
    r'"[a-zA-Z_][a-zA-Z0-9_](\.[a-zA-Z_][a-zA-Z0-9_])*"'
    return t

def t_IDENTIFIER(t):
    r'[a-zA-Z_][a-zA-Z0-9_]*'
    if t.value.lower() == 'true':
        t.type = 'BOOL'
        t.value = True
    elif t.value.lower() == 'false':
        t.type = 'BOOL'
        t.value = False
    else:
        t.type = reserved_words.get(t.value, 'IDENTIFIER')
    return t
    
def t_newline(t):
    r'\n+'
    t.lexer.lineno += len(t.value)

t_ignore  = ' \t'

def t_error(t):
    print "Illegal character '%s' on line %d, position %d." % \
        (t.value[0], t.lexer.lineno, t.lexer.lexpos)
    raise SystemExit()

# Build the lexer
lexer = lex.lex()



"""Chisels

Chisels are just functions that take a line and a buffer keep an eye out
for something interesting. If they don't find anything, they should just
return `(None, None)`

If, however, they do find something worth looking at, they continue down
the chain of filters until they exhaust them, or stop being interested
in them. 

Chisels can contain state too. The easiest way to do this is to create a 
closure, but if you'd rather think in objects, just override `__call__`.
"""

from canoe.filters.simple import *




from optparse import OptionParser

from config import Config
from watch import start_watch

parser = OptionParser()
parser.add_option('-c', '--config', default='', dest='config',
                  help='path to config file')

def main():
    (options, args) = parser.parse_args()
    conf = Config.from_file(options.config)
    if not conf:
        raise SystemExit()

    start_watch(conf)



from importlib import import_module

from canoe.canoe import Canoe
from canoe.routes import EchoRoute
from canoe.filters import PassThroughFilter

# TODO: need to import stuff in here I'm sure!

class Node(object):

    def evalute(self, env):
        raise NotImplementedError("Override me please")


class ConfigNode(Node):
    
    def __init__(self, top_assignments=None, watches=None):
        self.top_assignments = top_assignments
        self.watches = watches

    def evaluate(self, env):
        from config import Config
        if self.top_assignments:
            for x in self.top_assignments:
                x.evaluate(env)

        watches = [w.evaluate(env) for w in self.watches]
        return Config(watches)


class AssNode(Node):
    
    def __init__(self, identifier=None, value=None):
        self.identifier = identifier
        self.value = value

    def evaluate(self, env):
        if self.value:
            env[self.identifier] = self.value.evaluate(env)
        else:
            raise ValueError("Can't evaluated assignment without an rvalue")


class DeclNode(Node):
                   
    def __init__(self, typ, path=None, assignments=None):
        self.decl_type = typ
        self.path = path
        self.assignments = assignments

    def evaluate(self, env):
        kwargs = {}

        if self.assignments:
            for a in self.assignments:
                a.evaluate(kwargs)

        if self.decl_type == 'filter':
            klass = find_filter(self.path)
        elif self.decl_type == 'route':
            klass = find_route(self.path)
        else:
            raise ValueError("Don't know what to do with '%s'" % \
                                 self.decl_type)
        return klass(**kwargs)


class WatchNode(Node):

    def __init__(self, f, filters=None, route=None):
        self.file = f
        self.filters = filters
        self.route = route

    def evaluate(self, env):
        eroute = None
        efilters = None
        if self.route:
            eroute = self.route.evaluate(env)
        else:
            eroute = EchoRoute()

        if self.filters:
            efilters = [c.evaluate(env) for c in self.filters]
        else:
            efilters = [PassThroughChisel()]
            
        canoe = Canoe(efilters, eroute)

        # Let someone else deal with the fact that there can be
        # multiple canoes per file
        return (self.file, [canoe])


class ValueNode(Node):
    
    def __init__(self, value):
        print "Creating value node: ", value
        self.value = value

    def evaluate(self, env):
        if isinstance(self.value, DerefNode):
            return self.value.evaluate(env)
        return self.value


class DerefNode(Node):
    
    def __init__(self, identifier):
        self.identifier = identifier

    def evaluate(self, env):
        if self.identifier in env:
            return env[self.identifier]
        raise NameError("can't dereference '%s'. Not defined." % \
                            self.identifier)


# TODO: this is crap, and should definitely be smarter, allow for
# better pluginy infrastructure, etc
def find_filter(p):
    if '.' not in p:
        p = 'canoe.filters.' + p
    return _import_class(p)

def find_route(p):
    if '.' not in p:
        p = 'canoe.routes.' + p
    return _import_class(p)

def _import_class(p):
    bits = p.split('.')
    mod = '.'.join(bits[0:-1])
    cls = bits[-1]

    imod = import_module(mod)
    return getattr(imod, cls)



import ply.yacc as yacc

from lex import tokens
from ast import ConfigNode, AssNode, DeclNode, WatchNode, ValueNode, DerefNode

    
def p_config(p):
    """config : top_assignments watch_blocks
    """
    print "Config"
    p[0] = ConfigNode(top_assignments=p[1], watches=p[2])

def p_top_assignments(p):
    """top_assignments : top_assignment top_assignments
                       |
    """
    if len(p) == 1:
        p[0] = []
    else:
        p[0] = [p[1]] + p[2]

def p_watch_blocks(p):
    """watch_blocks : watch_block watch_blocks
                    |
    """
    if len(p) == 1:
        p[0] = []
    else:
        p[0] = [p[1]] + p[2]

def p_watch_block(p):
    """watch_block : WATCH STRING LCURLY filter_declarations \
                     route_declaration RCURLY
                     
    """
    p[0] = WatchNode(p[2], p[4], p[5])

def p_top_assignment(p):
    """top_assignment : IDENTIFIER EQUALS declaration
    """
    p[0] = AssNode(identifier=p[1], value=p[3])

def p_declaration(p):
    """declaration : filter_declaration
                   | route_declaration
    """
    p[0] = p[1]

def p_filter_declarations(p):
    """filter_declarations : filter_declaration filter_declarations
                           |
    """
    if len(p) == 1:
        p[0] = []
    else:
        p[0] = [p[1]] + p[2]

def p_filter_declaration(p):
    """filter_declaration : FILTER path LCURLY assignments RCURLY
                          | FILTER dereference
    """
    if len(p) == 3:
        p[0] = p[2]
    else:
        p[0] = DeclNode('filter', p[2], p[4])

def p_route_declaration(p):
    """route_declaration : ROUTE path LCURLY assignments RCURLY
                         | ROUTE dereference
    """
    if len(p) == 3:
        p[0] = p[2]
    else:
        p[0] = DeclNode('route', p[2], p[4])

def p_assignments(p):
    """assignments : assignment COMMA assignments
                   | assignment
    """
    if len(p) == 2:
        p[0] = [p[1]]
    else:
        p[0] = [p[1]] + p[3]

def p_assignment(p):
    """assignment : IDENTIFIER EQUALS value
    """
    p[0] = AssNode(p[1], p[3])

def p_path(p):
    """path : IDENTIFIER
            | IMPORT_PATH
    """
    p[0] = str(p[1])

def p_value(p):
    """value : NUMBER 
             | STRING
             | BOOL
             | dereference
    """
    possible_values = {
        'false': False,
        'False': False,
        'true': True,
        'True': True,
        }
    p[0] = ValueNode(possible_values.get(p[1], p[1]))

def p_dereference(p):
    """dereference : AT IDENTIFIER
    """
    p[0] = DerefNode(p[2])

def p_error(p):
    raise SyntaxError("Syntax Error! on line %d" % p.lexer.lineno)


parser = yacc.yacc()




def lift_iter(xs):
    if not xs:
        return []
    if hasattr(xs, '__iter__'):
        return xs
    return [xs]


class Canoe(object):

    def __init__(self, filters=None, route=None):
        self._filters = lift_iter(filters)
        self._route = route

    def send(self, line, buffer):
        for f in self._filters:
            nline, nbuffer = f(line, buffer)
            if nline is None or nbuffer is None:
                break

        if nline != None and nbuffer != None:
            if self._route:
                return self._route(line, buffer)

    def add_filter(self, filter):
        self._filters.append(filters)

    def set_route(self, route):
        self._route = route



import os
import sys
import time
import threading

from data import Buffer                  
      
class WatchedFile(object):
    
    def __init__(self, fname, bufn=0):
        self.filename = fname
        self.fd = open(fname, 'r')
        self.st = os.stat(fname)
        self.buf = Buffer(bufn)
        self.ends_nl = False

    def _reopen(self):
        self.fd.close()
        self.fd = open(fname, 'r')

    def _chunk(self, chk, cb=None):
        lines = chk.split('\n')
        l = len(lines)
        if l > 0:
            last_line = None
            for i in xrange(l):
                if i == 0 and not self.ends_nl:
                    last_line = self.buf.alter(lambda x: \
                                                   (x or '') + lines[0])
                elif i == (l - 1):
                    last_line = lines[-1]
                    if last_line.endswith('\n'):
                        self.ends_nl = True
                else:
                    last_line = lines[i]

                self.buf.push(last_line)
                if cb:
                    cb(last_line, self.buf)

    def watch(self, cb=None):
        # TODO fill the buffer with bufn lines
        # TODO this doesn't work with unix sockets, 
        # nor does it work with fifos, since seek doesn't work on them
        self.fd.seek(0, 2)

        lastTell = -1

        while True:
            a = self.fd.read()
            if self.filename:
                tell = self.fd.tell()
            else:
                tell += len(a)

            if tell > lastTell:
                self._chunk(a, cb)

            lastTell = tell

            if self.filename:
                st = os.stat(self.filename)
                if st.st_dev != self.st.st_dev or \
                        st.st_ino != self.st.st_ino or \
                        st.st_nlink != self.st.st_nlink:
                    self.fd.close()
                    self._reopen()
                    lastTell = -1
                time.sleep(.1)


def start_watch(conf):
    """Given a config object which may define multiple canoes, run
    all of them, however that means.
    """
    def mk_cball(canoes):
        def cb(line, buffer):
            for c in canoes:
                c.send(line, buffer)
        return cb

    ts = []

    for fname, canoes in conf.watching:
        wf = WatchedFile(fname, 1024) # TODO: should be configured
        cb = mk_cball(canoes)
        tid = threading.Thread(target=wf.watch, args=(cb,))
        ts.append(tid)
        tid.start()

    for t in ts:
        t.join()




from setuptools import setup, find_packages

setup(
    name = "canoe",
    version = "0.1.1",
    packages = find_packages(),
    entry_points = {
        'console_scripts': [
            'canoe = canoe.main:main',
        ],
        },
    author = "Andrew Gwozdziewycz",
    author_email = "web@apgwoz.com",
    description = "Carve out log files and paddle",
    install_requires = ['ply'],
    license = "GPL",
    keywords = "log analyze alert monitor",
    url = "https://github.com/apgwoz/canoe",
)



import smtplib
import rfc822

from email.mime.text import MIMEText


class EchoRoute(object):
    
    def __init__(self, prefix=None):
        self._prefix = prefix or ''
    
    def __call__(self, line, buffer):
        print "%s%s" % (self._prefix, line)


def quote_addrs(addrs):
    return map(smtplib.quoteaddr, addrs)


def sanitize_addrs(addrs):
    naddrs = []
    for addr in addrs:
        _, a = rfc822.parseaddr(addr.strip())
        if a and len(a) > 0:
            naddrs.append(a)
        else:
            print "ERROR: ", _, a

    return quote_addrs(naddrs)


class SMTPRoute(object):

    def __init__(self, **kwargs):
        self._to = map(lambda x: x.strip(), 
                       kwargs.pop('toaddr', '').split(','))
        self._from = kwargs.pop('fromaddr', 'Canoe <noreply@localhost>')
        self._subject = kwargs.pop('subject', 'Message from canoe')
        self._buffer_lines = kwargs.pop('buffer_lines')
        self._testing = kwargs.pop('testing', False)
        self._smtp_host = kwargs.pop('smtp_host')
        self._smtp_port = kwargs.pop('smtp_port')
        self._smtp_login = kwargs.pop('smtp_login')
        self._smtp_password = kwargs.pop('smtp_password')

    def __call__(self, line, buffer):
        tos = sanitize_addrs(self._to)
        bt = buffer.copyn(20)
        text = '\r\n'.join(bt) + '\r\n' + line
        msg = MIMEText(text)
        msg['Subject'] = self._subject
        msg['To'] = ', '.join(tos)
        msg['From'] = ', '.join(sanitize_addrs([self._from]))

        if self._testing:
            # TODO: send this of course
            print msg.as_string()
            print "-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=\n\n\n"
        else:
            smtp = smtplib.SMTP(self._smtp_host, self._smtp_port)
            smtp.login(self._smtp_login, self._smtp_password)
            smtp.sendmail(msg['From'], tos, msg.as_string())
            smtp.quit()





class Buffer(object):
    
    class EmptyValue:
        pass

    def __init__(self, n):
        self.n = n + 1
        self.start = 0
        self.end = 0
        self.els = [None for n in range(n + 1)]

    def full(self):
        return (self.end + 1) % self.n == self.start

    def empty(self):
        return self.end == self.start
        
    def push(self, val):
        self.els[self.end] = val
        self.end = (self.end + 1) % self.n
        if self.end == self.start:
            self.start = (self.start + 1) % self.n

    def alter(self, fn):
        tmp = fn(self.els[self.start])
        if not self.empty():
            self.els[self.start] = tmp
        else:
            self.push(tmp)
        return tmp

    def pop(self):
        if self.empty():
            raise ValueError("queue is empty")
        elem = self.els[self.start]
        self.els[self.start] = None
        self.start = (self.start + 1) % self.n
        return elem

    def copyn(self, n):
        """Copies last n items out of the buffer"""
        out = []

        if n >= self.n:
            n = self.n

        i = 0
        e = (self.end - 1) % self.n

        while e != self.start and i < n:
            out.append(self.els[e]) 
            e = (e - 1) % self.n
            i += 1

        out.reverse()
        return out

